/* light_switch
**     example use of GtkSwitch (new in gtk-3 )
**
**    Copyright (C) 2011    John Lumby    johnlumby@hotmail.com
**    This program is based on,  and uses verbatim some parts of,
**    the GTK tutorial at http://library.gnome.org/devel/gtk-tutorial/
**    and several sample programs provided by the
**    Gtk and Gnome organizations,  www.gtk.org, www.gnome.org,
**    who may also hold copyrights:
**    You should consult the current reference documentation available at
**              http://www.gtk.org/documentation.html
**              http://library.gnome.org/devel/references
**    for details of how to write programs such as this.
**
**
*--------------------------------------------------------------------------*
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  The GNU General Public License, version 2, is viewable at
**              http://www.gnu.org/licenses/gpl-2.0.html
**  The following is a brief summary of specific points from the license
**  that apply to use of this program,  but not a replacement of that license.
**  Permission is granted to use this program for any legal purpose and to copy,
**  modify, or redistribute the program subject to the terms and restrictions
**  stated in the GNU General Public License, version 2
**  and specifically provided that you abide by both of the following
**  as well as that license :
**   .   You may not incorporate this program into a proprietary program either in part
**       or in whole,  either in source or compiled form,   proprietary meaning
**       either that you claim ownership of the program or that you sell it
**       for a fee or both.
**   .   You must retain the above copyright statement and this entire notice
**       in any copy or redistribution,  and make it clear that the licensing terms,
**       restrictions and obligations apply to whomever you give or redistribute
**       this program to, including this requirement to pass on the same terms,
**       restrictions and obligations to that recipient.
*--------------------------------------------------------------------------*
**
**
** light_switch provides an on/off mechanism.
** The user can control which state should be active
** by clicking the empty area or by dragging the handle.
**
** If compiled against gtk-2 or compile flag -DUSE_BUTTON,  use a toggle button instead
**
**  how to run it :    see usage below
**
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <glib.h>
#include <pango/pangocairo.h>

GtkStyle *gtk_widget_get_style (GtkWidget *widget);
gchar   **gtk_rc_get_default_files ();

char usage_text[] = " light_switch : example use of GtkSwitch\n"
  "                   to turn a light image on and off.\n"
  "    invoke as\n"
  "       light_switch [-h] [-x signal_index ] [-l] [-s] [-v]\n"
  "     signal_index may be 0 1 or 2 to select which of the following signals to connect to :\n"
  "                         0 event-after\n"
  "                         1 notify::active\n"
  "                         2 toggled\n"
  "     -l     requests a list of all available pango fonts\n"
  "            which is useful when experimenting with setting different fonts via settings and styles\n"
  "     -s     requests a list of sources of settings files (works only for GTK2)\n"
  "     -v     verbose (maybe)\n";

static int verbose = 0;
static GtkEntry *entrybox;

static void
list_fonts ()
{
    int i;
    PangoFontFamily ** families;
    int n_families;
    PangoFontMap * fontmap;

    fontmap = pango_cairo_font_map_get_default();
    pango_font_map_list_families (fontmap, & families, & n_families);
    printf ("There are %d families\n", n_families);
    for (i = 0; i < n_families; i++) {
        PangoFontFamily * family = families[i];
        const char * family_name;

        family_name = pango_font_family_get_name (family);
        printf ("Family %d: %s\n", i, family_name);
    }
    g_free (families);
}

/*   we could base test on #if (GTK_MAJOR_VERSION < 3) but this is more specific to whether the GTkSwitch widget is available */
#if (!defined (__GTK_SWITCH_H__ )) || defined(USE_BUTTON)  /*   if on gtk-2,  redefine switch objects as togglebuttons */
#ifndef USE_BUTTON
#define USE_BUTTON
#endif  /* ndef USE_BUTTON */
#define GtkSwitch GtkToggleButton
#define gtk_switch_new gtk_toggle_button_new_with_label
#define gtk_switch_get_active gtk_toggle_button_get_active
#define gtk_switch_set_active gtk_toggle_button_set_active
#define switch_label "Toggle Me"
#define WINDOW_TITLE "ToggleSwitch"
#define DEFAULT_SIGNAL_INDEX 2 /* toggled */
#else
#define switch_label
#define WINDOW_TITLE "LightSwitch"
#define DEFAULT_SIGNAL_INDEX 0 /* event-after */
#endif /*  on gtk-2  */

/*  some callbacks interpose a paramspec parameter between objptr and dataptr ! */
static void
turn_light_on_off (GtkSwitch *light_switch_widg,  GParamSpec *arg1,  GtkWidget *light_img_widg)
{

  gboolean on_off = gtk_switch_get_active (light_switch_widg);
#ifdef DEBUG_LIGHT_SWITCH
  g_print ("turn_light_on_off: Light is %d dataptr= 0x%X widget is %d\n"
    , on_off , (unsigned int)light_img_widg , GTK_IS_WIDGET (light_img_widg) );
#endif /*  DEBUG_LIGHT_SWITCH */

  gtk_widget_set_sensitive ( light_img_widg, on_off);

  if (on_off)
     gtk_entry_set_text  (entrybox, "we are wide awake");
  else
     gtk_entry_set_text  (entrybox, "we are fast asleep");
  
}

static void
toggle_light_on_off (GtkSwitch *light_switch_widg, GtkWidget *light_img_widg)
{
  turn_light_on_off  (light_switch_widg,  0,  light_img_widg);
}

GtkWidget *
do_light_switch (char *signal_name)
{
  GtkWidget *window = NULL;
  GtkImage *light_on_image;

  GtkWidget *vbox;
  GtkWidget *hbox;

  GtkSwitch *light_switch_widg;
  GtkWidget *content_area;
#ifndef USE_BUTTON  /*   if on gtk-3 */
  gint slider_width;
  GValue value = {0};
#endif /*  ndef USE_BUTTON  if on gtk-3 */

  GtkWidget *tbox;
  GtkStyle *entry_box_style;  /*  style of the entrybox */
  const char *entry_font_description;  /*  font of the entry box  */
  GtkStyle *switch_style;  /*  style of the switch/toggle widget      */
  const char *switch_font_description;  /*  font of the switch/toggle */

  GtkSettings *default_gtk_settings;
  gchar *cur_theme_str;

      window = gtk_dialog_new_with_buttons ("GtkSwitch",
			  0,
			  0,
			  GTK_STOCK_CLOSE,
			  GTK_RESPONSE_NONE,
			  NULL);
      gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
      gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE);
      
      g_signal_connect (window, "response",
			G_CALLBACK (gtk_widget_destroy), NULL);
      g_signal_connect (window, "destroy",
			G_CALLBACK (gtk_widget_destroyed), &window);

      content_area = gtk_dialog_get_content_area (GTK_DIALOG (window));

      vbox = gtk_vbox_new (GTK_ORIENTATION_VERTICAL, 85);
      gtk_box_pack_start (GTK_BOX (content_area), vbox, TRUE, TRUE, 0);
      gtk_container_set_border_width (GTK_CONTAINER (vbox), 8);

      light_switch_widg = (GtkSwitch *)gtk_switch_new (switch_label);

      gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (light_switch_widg), FALSE, FALSE, 0);
      
      hbox = gtk_hbox_new (GTK_ORIENTATION_HORIZONTAL, 85);
      gtk_box_pack_end (GTK_BOX (content_area), hbox, TRUE, TRUE, 0);
      gtk_container_set_border_width (GTK_CONTAINER (hbox), 8);

      light_on_image = GTK_IMAGE (gtk_image_new_from_file ("light_on.png"));

      gtk_switch_set_active (light_switch_widg , FALSE);
      gtk_box_pack_end (GTK_BOX (hbox), GTK_WIDGET (light_on_image ), FALSE, FALSE, 0);


      /*  It is difficult to know how to persuade gtk to run the callback for my Switch widget
      **  because there is no signal specific to a GtkSwitch.     Two signal-types do what we need:
      **  event-after ,  notify::active
      **  if on gtk-2 and the widget is a togglebutton,
      **  then the provided signal is "toggled",  which takes a dyadic function
      */

     g_signal_connect_object ( GTK_WIDGET (light_switch_widg), signal_name
                        , (strcmp(signal_name,"toggled") ? G_CALLBACK (turn_light_on_off) : G_CALLBACK (toggle_light_on_off))
                        , GTK_WIDGET (light_on_image) , G_CONNECT_AFTER);

     gtk_widget_set_sensitive ( GTK_WIDGET (light_on_image), FALSE);

#ifndef USE_BUTTON  /*   if on gtk-3 */
     g_value_init (&value, G_TYPE_INT);
     gtk_widget_style_get_property ( GTK_WIDGET (light_switch_widg),"slider-width",&value);
     slider_width = g_value_get_int (&value);
     if (verbose)
       fprintf (stderr,"do_light_switch : slider-width = %d\n",slider_width);
#endif /*  ndef USE_BUTTON   if on gtk-3 */

     entrybox = (GtkEntry *)gtk_entry_new();
     gtk_entry_set_text  (entrybox, "we are fast asleep");
     tbox = gtk_hbox_new (GTK_ORIENTATION_HORIZONTAL, 85);
     gtk_box_pack_end (GTK_BOX (content_area), tbox, TRUE, TRUE, 0);
     gtk_container_set_border_width (GTK_CONTAINER (tbox), 8);
     gtk_box_pack_start (GTK_BOX (tbox), GTK_WIDGET (entrybox), FALSE, FALSE, 0);

     default_gtk_settings = gtk_settings_get_default();
     g_object_get (default_gtk_settings, "gtk-theme-name", &cur_theme_str, NULL);
     if (verbose)
       fprintf (stderr,"do_light_switch : current-theme is %s\n",cur_theme_str);
     g_free (cur_theme_str);

#ifdef DEBUG_LIGHT_SWITCH
     g_print ("do_light_switch : Light is %d dataptr= 0x%X widget is %d\n"
      , gtk_switch_get_active (light_switch_widg) , (unsigned int) GTK_WIDGET (light_on_image) , GTK_IS_WIDGET (light_on_image) );
#endif /*  DEBUG_LIGHT_SWITCH */

     /*  report on the font of the style attached to the entrybox
     **  in order to understand whether any particular rc file or settings or css had any effect
     */
     entry_box_style = gtk_widget_get_style ( GTK_WIDGET (entrybox) );
     if (entry_box_style)
       entry_font_description = pango_font_description_to_string (entry_box_style->font_desc);
     else
       entry_font_description = "unknown";

     switch_style = gtk_widget_get_style ( GTK_WIDGET ( light_switch_widg ) );
     if (switch_style)
       switch_font_description = pango_font_description_to_string (switch_style->font_desc);
     else
       switch_font_description = "unknown";
     if (verbose)
       fprintf (stderr,"entry_box font is %s   "
#ifndef USE_BUTTON  /*   if on gtk-3 */
         "switch"
#else  /* ndef USE_BUTTON */
         "toggle"
#endif    /* def USE_BUTTON */
              " font is %s\n",entry_font_description,switch_font_description);

  if (!gtk_widget_get_visible (window))
    gtk_widget_show_all (window);
  else
    gtk_widget_destroy (window);

  return window;
}

int
main (int argc, char **argv)
{
  GtkWidget *switch_window;

  int option;  /*  returned by getopt */
  extern int optind, opterr, optopt;
  extern char *optarg;
  int rc = 0;
  char signal_names[3][32] = { { 'e' , 'v' , 'e' , 'n' , 't' , '-' , 'a' , 'f' , 't' , 'e' , 'r' , '\0' }
                             , { 'n' , 'o' , 't' , 'i' , 'f' , 'y' , ':' , ':' , 'a' , 'c' , 't' , 'i' , 'v' , 'e' , '\0' }
                             , { 't' , 'o' , 'g' , 'g' , 'l' , 'e' , 'd' , '\0' } };
  unsigned int signal_index = DEFAULT_SIGNAL_INDEX;
  gchar **default_files_array;

  gtk_init (&argc, &argv);

  while ( (option = getopt (argc, argv, ":hlsx:v")) != -1)
  {
#ifdef DEBUG_OPTION
    fprintf(stderr,"option %c\n",option);
#endif
    switch (option)
    {
    case 'h':
      fprintf(stderr,"%s\n",usage_text);
      exit (0);
    case 'l':
      list_fonts ();
      break;
    case 's':
      default_files_array = gtk_rc_get_default_files();
      if (default_files_array) {
        while (*default_files_array) {
          fprintf (stderr,"source of default settings : %s\n",*default_files_array);
          default_files_array++;
        }
      }
      break;
    case 'x':
      if ((*optarg < '0') || (*optarg > '9')) {
        fprintf(stderr,"%s\n",usage_text);
        exit (0);
      }
      signal_index = atoi(optarg);
      break;
    case 'v':
      verbose++;
      break;

    case '?':
      rc = -1;
      break;

    default:
      fprintf(stderr, "ERROR: getopt returned character code 0%o for option '%c' unknown.\n"
               ,option ,*(argv[optind] + 1));
      rc = -1;
    }
  }

  if (verbose)
    fprintf (stderr,"using signal index %d %s\n",signal_index,signal_names[signal_index]);

  switch_window = do_light_switch (signal_names[signal_index]);

  g_signal_connect_after (switch_window, "destroy",
	    G_CALLBACK (gtk_main_quit), NULL);

  gtk_main ();

  return rc;
}
