/*    watch_an_app : example use of GtkApplication to watch for 
**                   other processes to start and stop.
**
**    Copyright (C) 2011    John Lumby    johnlumby@hotmail.com
**    This program is based on,  and uses verbatim some parts of,
**    the following sample programs provided by the
**    Gtk and Gnome organizations,  www.gtk.org, www.gnome.org,
**    who may also hold copyrights:
**     GtkApplication Example 110. A simple application
**     GApplication   Example 14. A GApplication with actions
**     Gio Convenience API for watching bus names
**                    Example 9. Simple application watching a name
**    You should consult the current reference documentation available at
**              http://www.gtk.org/documentation.html
**              http://library.gnome.org/devel/references
**    for details of how to write programs such as this.
**
**
*--------------------------------------------------------------------------*
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  The GNU General Public License, version 2, is viewable at
**              http://www.gnu.org/licenses/gpl-2.0.html
**  The following is a brief summary of specific points from the license
**  that apply to use of this program,  but not a replacement of that license.
**  Permission is granted to use this program for any legal purpose and to copy,
**  modify, or redistribute the program subject to the terms and restrictions
**  stated in the GNU General Public License, version 2
**  and specifically provided that you abide by both of the following
**  as well as that license :
**   .   You may not incorporate this program into a proprietary program either in part
**       or in whole,  either in source or compiled form,   proprietary meaning
**       either that you claim ownership of the program or that you sell it
**       for a fee or both.
**   .   You must retain the above copyright statement and this entire notice
**       in any copy or redistribution,  and make it clear that the licensing terms,
**       restrictions and obligations apply to whomever you give or redistribute
**       this program to, including this requirement to pass on the same terms,
**       restrictions and obligations to that recipient.
*--------------------------------------------------------------------------*
**
**  The program registers itself as an application using the supplied connection name,
**  and if the user specifies a name to watch,   sets a watch on that name.
**  It creates a statusbar which displays what it is doing
**  and whether the watched-for application is active.
**
**  how to run it :    see usage below
**
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gio/gio.h>
#include <gtk/gtk.h>

#include <glib.h>
#include <pango/pangocairo.h>

char usage_text[] = " watch_an_app : example use of GtkApplication\n"
  "                   to watch for other processes to start and stop.\n"
  "    invoke as\n"
  "      watch_an_app -n my_connection_name [ -w watched_connection_name ] [-h] [-v]\n"
  "    -n my_connection_name      the connection name (sometimes called bus name)\n"
  "                               of this process, and is mandatory\n"
  "    -w watched_connection_name the connection name which this process watches for\n"
  "                                               which is optional\n"
  "    Note - a connection name can be any string that contains one or more dots.\n"
  "    -h   this help\n"
  "    -v   verbose output - (maybe)\n"
  "    program output goes to stderr\n";

static gchar my_connection_name[256] = { '\0' } ;/*  my unique connection name aka Bus Name */
static gchar watched_connection_name[256] = { '\0' } ;/*  connection name to watch */
static gchar watching_message[384] = { 'k' , 'e' , 'e' , 'p' , 'i' , 'n' , 'g' , ' ' , 'a' , ' ' , 's' , 'h' , 'a' , 'r' , 'p' , ' ' , 'l' , 'o' , 'o' , 'k' , '-' , 'o' , 'u' , 't' , ' ' , 'f' , 'o' , 'r' , ' ' , '\0' } ;/* message in statusbar */
static int verbose = 0;
#define MY_CONTEXT_DESCRIPTION "watching"
#define WINDOW_WIDTH_IN_PIXELS 456
#define WINDOW_HEIGHT_IN_PIXELS 226

static void
on_name_appeared (GDBusConnection *connection,
                  const gchar     *name,
                  const gchar     *name_owner,
                  gpointer         user_data)
{
  guint context_id;
  char watchmessage[256];
  GtkWidget *statusbar = GTK_WIDGET(user_data);

  context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR(statusbar), MY_CONTEXT_DESCRIPTION);
  gtk_statusbar_pop  (GTK_STATUSBAR(statusbar), context_id);
  sprintf(watchmessage,"Name %.96s is owned by %.96s", name, name_owner);
  gtk_statusbar_push (GTK_STATUSBAR(statusbar), context_id, watchmessage);
}

static void
on_name_vanished (GDBusConnection *connection,
                  const gchar     *name,
                  gpointer         user_data)
{
  guint context_id;
  GtkWidget *statusbar = GTK_WIDGET(user_data);

  context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR(statusbar), MY_CONTEXT_DESCRIPTION);
  gtk_statusbar_pop  (GTK_STATUSBAR(statusbar), context_id);
  gtk_statusbar_push  (GTK_STATUSBAR(statusbar), context_id, watching_message);
}

static void
activate (GApplication *application)
{
  if (verbose)
    g_print ("activated\n");
}

static void
propschg (GApplication *application)
{
  if (verbose)
    g_print ("Properties Changed\n");
}

static void
activate_action (GAction  *action,
                 GVariant *parameter,
                 gpointer  data)
{
  if (verbose)
    g_print ("action %s activated\n", g_action_get_name (action));
}

static void
add_actions (GApplication *app)
{
  GSimpleActionGroup *actions;
  GSimpleAction *action;

  actions = g_simple_action_group_new ();

  action = g_simple_action_new ("simple-action", NULL);
  g_signal_connect (action, "activate", G_CALLBACK (activate_action), app);
  g_simple_action_group_insert (actions, G_ACTION (action));
  g_object_unref (action);

  g_application_set_action_group (app, G_ACTION_GROUP (actions));
  g_object_unref (actions);
}

/*  function  to return dimensions in pixels of text in specified font
*   copied from mailing list post http://ubuntuforums.org/showthread.php?t=1248590
*/
void gettextwdht ( PangoContext *widget_pango_context,
                   char * stringtomeasure ,
                   int * wdret , int * htret )
{
   PangoFontDescription * fd;
   PangoLayout * layout;

   fd = pango_context_get_font_description(widget_pango_context);
   layout = pango_layout_new ( widget_pango_context );
   pango_layout_set_text ( layout, stringtomeasure, -1 );
   pango_layout_set_font_description ( layout, fd );
   pango_layout_get_pixel_size (layout, wdret , htret );
   g_object_unref ( layout );
}

static void
new_window (GApplication *app,
            GFile        *file)
{
  GtkWidget *window, *statusbar;
  guint watcher_id;
  guint context_id;
  gint desired_window_width;
  gint desired_window_height;
  PangoContext *statusbar_pango_context;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_application (GTK_WINDOW (window), GTK_APPLICATION (app));
  gtk_window_set_title (GTK_WINDOW (window), my_connection_name);

  statusbar = gtk_statusbar_new ();
  /*     In a statusbar, text that exceeds width of containing window is truncated to ellipsis.
  **     If that were to happen,  user would have to expand the window manually to reveal truncated part.
  **     So,  here we need to obtain the width of the text in the statusbar in pixels,  to set window width
  */
  statusbar_pango_context = gtk_widget_get_pango_context(statusbar);
  gettextwdht ( statusbar_pango_context , watching_message , &desired_window_width , &desired_window_height);
  /*  empirically,   we probably want the window dimensions to be about 64 pixels greater than size of text */
  desired_window_width += 64;
  desired_window_height += 64;
  gtk_window_set_default_size (GTK_WINDOW (window), desired_window_width , desired_window_height);

  context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR(statusbar), MY_CONTEXT_DESCRIPTION);
  gtk_statusbar_push  (GTK_STATUSBAR(statusbar), context_id, watching_message);
  gtk_container_add (GTK_CONTAINER (window), statusbar);
  gtk_widget_show (statusbar);

  gtk_widget_show_all (GTK_WIDGET (window));

  if (*watched_connection_name != '\0') {
    watcher_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
                                 watched_connection_name,
                                 G_BUS_NAME_WATCHER_FLAGS_NONE,
                                 on_name_appeared,
                                 on_name_vanished,
                                 (gpointer)statusbar,
                                 NULL);
  }
}

static void
watch_an_app_activate (GApplication *application)
{
  new_window (application, NULL);
}

static void
watch_an_app_open (GApplication  *application,
                GFile        **files,
                gint           n_files,
                const gchar   *hint)
{
      new_window (application, NULL);
}

typedef GtkApplication WatchApp;
typedef GtkApplicationClass WatchAppClass;

G_DEFINE_TYPE (WatchApp, watch_an_app, GTK_TYPE_APPLICATION)

static void
watch_an_app_finalize (GObject *object)
{
  G_OBJECT_CLASS (watch_an_app_parent_class)->finalize (object);
}

static void
watch_an_app_init (WatchApp *app)
{
}

static void
watch_an_app_class_init (WatchAppClass *class)
{
  G_OBJECT_CLASS (class)->finalize= watch_an_app_finalize;

  G_APPLICATION_CLASS (class)->activate = watch_an_app_activate;
  G_APPLICATION_CLASS (class)->open = watch_an_app_open;
}

WatchApp *
watch_an_app_new (void)
{

  WatchApp *nice_new_app_watcher;

  g_type_init ();

  nice_new_app_watcher = g_object_new (watch_an_app_get_type (),
                       "application-id", my_connection_name,
                       "flags", G_APPLICATION_HANDLES_OPEN,
                       NULL);

  g_signal_connect (nice_new_app_watcher, "activate", G_CALLBACK (activate), NULL);
  g_signal_connect (nice_new_app_watcher, "notify::active", G_CALLBACK (propschg), NULL);
  /*  timeout 1 secs after last g_application_release(),  which occurs after closing window */
  g_application_set_inactivity_timeout ((GApplication *)nice_new_app_watcher, 1000);

  add_actions ((GApplication *)nice_new_app_watcher);

  g_application_register ((GApplication *)nice_new_app_watcher, NULL, NULL);
  g_action_group_activate_action (G_ACTION_GROUP (nice_new_app_watcher), "simple-action", NULL);

  return nice_new_app_watcher;

}

int
main (int argc, char **argv)
{
  WatchApp *watch_an_app;
  int status;
  char my_file_path[128] = { '\0' };

  int option;  /*  returned by getopt */
  extern int optind, opterr, optopt;
  int rc = 0;

  while ( (option = getopt (argc, argv, ":hn:w:v")) != -1)
  {
#ifdef DEBUG_OPTION
    fprintf(stderr,"option %c\n",option);
#endif
    switch (option)
    {
    case 'h':
      fprintf(stderr,"%s\n",usage_text);
      exit (0);
    case 'n':
      strncpy(my_connection_name, optarg, (sizeof(my_connection_name)-1));
      break;
    case 'w':
      strncpy(watched_connection_name, optarg, (sizeof(watched_connection_name)-1));
      break;
    case 'v':
      verbose++;
      break;

    case '?':
      rc = -1;
      break;

    default:
      fprintf(stderr, "ERROR: getopt returned character code 0%o for option '%c' unknown.\n"
               ,option ,*(argv[optind] + 1));
      rc = -1;
    }
  }

  if (*my_connection_name == '\0') {
      fprintf(stderr,"\nyou must specify -n my_connection_name,\notherwise I don't know who I am\n");
      fprintf(stderr,"%s\n",usage_text);
      exit(9);
  }
  else if (!strchr(my_connection_name,'.')) {
      fprintf(stderr,"\nyou must include at least one dot in my_connection_name,\notherwise D-Bus doesn't know what to do\n");
      fprintf(stderr,"%s\n",usage_text);
      exit(11);
  }

  if (*watched_connection_name != '\0') {
      if (!strchr(watched_connection_name,'.')) {
        fprintf(stderr,"\nyou must include at least one dot in watched_connection_name,\notherwise D-Bus can't find it\n");
        fprintf(stderr,"%s\n",usage_text);
        exit(13);
      }
      strncat(watching_message, watched_connection_name 
           ,(sizeof(watching_message) - strlen(watching_message) - 1));
  } else {
    strncpy(watching_message, "not watching anyone"
           ,(sizeof(watching_message) - 1));
  }

  if (verbose)
    fprintf(stderr,"my connection name is %s   watched connection name is %s\n",my_connection_name ,watched_connection_name);

  if (argc > optind)
    strncpy(my_file_path, argv[optind++] , (sizeof(my_file_path)-1));

  watch_an_app = watch_an_app_new ();

  status = g_application_run (G_APPLICATION (watch_an_app), argc, argv);
  g_object_unref (watch_an_app);

  return status;
}
